# -*- coding: utf-8 -*-
import glob
import law
import luigi
import logging
import os
import re
import shutil
import six

from functools import partial
from fnmatch import fnmatch
from subprocess import PIPE

from law.util import colored

from analysis.tasks.FastWarm import FastWarm

from analysis.framework import Task
from analysis.mixins import HTCondorRemoteStorageMixin, TarballExtractionMixin
from analysis.util import interruptable_popen, working_directory

logger = logging.getLogger(__name__)
working_directory = partial(working_directory, logger=logger)


class MergeFastWarm(Task, TarballExtractionMixin, HTCondorRemoteStorageMixin):
    process = luigi.Parameter()
    observables = luigi.ListParameter()

    def requires(self):
        return FastWarm()

    def output(self):
        return self.remote_or_local_target(
            '{}.{}.fastwarm.tar.gz'.format(
                self.process,
                self.name,
            ),
        )

    def run(self):
        rundirpath = os.getcwd()
        tmpdirpath = os.path.join(rundirpath, 'tmpdir')

        logger.debug("Running job in directory: %s", rundirpath)
        job_ok = True  # set to False if exception is raised

        try:
            # -- create temporary directory
            if not os.path.exists(tmpdirpath):
                logger.debug("Creating temporary directory: %s", tmpdirpath)
                os.mkdir(tmpdirpath)
            elif os.path.isdir(tmpdirpath):
                logger.warning("Temporary directory already exists: %s", tmpdirpath)
            else:
                raise Exception("Could not create temporary directory: %s".format(tmpdirpath))

            # ensure output parent directory exists
            logger.debug("Creating output parent directory: %s", self.output().parent.path)
            self.output().parent.touch()

            # run all subsequent command in tmpdir
            with working_directory(tmpdirpath):
                inputs = self.input()['collection'].targets
                logger.debug("Merging %d input files.", len(inputs))
                for inp in six.itervalues(inputs):
                    logger.debug("Loading input file: %s", inp)
                    try:
                        # extract input tarball into tmpdir
                        self.extract_tarball(
                            inp,
                            law.LocalDirectoryTarget(tmpdirpath),
                            formatter="tar",
                            logger=logger,
                        )
                    except:
                        logger.warning("Input file does not exist: %s", inp)

                # Remove fastwarm logs and runcards before output preps
                logger.debug("Removing fastwarm logs and runcards.")
                os.system('rm *fastwarm*.log')
                os.system('rm *fastwarm*.run')

                for obs in self.observables:
                    logger.debug("Running `fnlo-add-warmup` for observable: %s", obs)
                    wrmfile = '{}.{}.wrm'.format(self.process, obs)
                    logfile = '{}.{}.addwarmup.log'.format(self.process, obs)
                    errfile = '{}.{}.addwarmup.err'.format(self.process, obs)
                    logger.debug("Wrm filename is: %s", wrmfile)
                    logger.debug("Log filename is: %s", logfile)

                    command_line_args = [
                        'fnlo-add-warmup_v23.pl',
                        '-d', '-v', '2.4', '-w', '.',
                        '-o', wrmfile, obs
                    ]
                    logger.info("Running command: %s",
                        colored(' '.join(command_line_args), 'yellow'))
                    code, out, error = interruptable_popen(
                        command_line_args,
                        stdout=PIPE,
                        stderr=PIPE
                    )

                    logger.debug("Return code: %d", code)
                    logger.debug("Standard output:\n%s\n", out)
                    logger.debug("Standard error:\n%s\n", error)

                    with open(logfile, 'w') as f:
                        f.write(out)
                    with open(errfile, 'w') as f:
                        f.write(error)

                    if (code == 0):
                        logger.info("NNLOJET status: SUCCESS")
                        job_ok = True
                    else:
                        raise RuntimeError("NNLOJET status: FAIL (exit code {})".format(code))

                    # Remove unmerged wrm files before output preps
                    os.system('rm {}.*.{}.s*.wrm'.format(self.process, obs))


        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            job_ok = False
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise

        # finalize job
        finally:
            logger.debug("Finalizing job...")

            with working_directory(tmpdirpath):
                logger.debug("Files in current directory:\n  %s", '\n  '.join(glob.glob('*')))

                # write out results on success, debug info on fail
                if job_ok:
                    tarfilter = lambda n: n if (
                        n.name == '.' or
                        fnmatch(n.name, '*.wrm') or fnmatch(n.name, '*.log')
                    ) else None
                    logger.debug(
                        "Job successful. Writing output tarfile from directory: %s", os.getcwd()
                    )
                    self.output().dump(os.getcwd(), mode="w:gz", formatter="tar", add_kwargs=dict(filter=tarfilter))
                else:
                    debugfile = re.sub(
                        '.tar.gz$', '.debug.tar.gz', os.path.basename(self.output().path)
                    )
                    logger.debug(
                        "Job failed. Writing debug tarfile: %s -> %s",
                        os.getcwd(), debugfile
                    )
                    self.remote_target(debugfile).dump(os.getcwd(), mode="w:gz", formatter="tar")

            logger.debug("Removing temporary directory: %s", tmpdirpath)
            shutil.rmtree(tmpdirpath)
