import logging
import os
import six
import subprocess

from contextlib import contextmanager
from law.util import colored

logger = logging.getLogger(__name__)


def pretty_print_command(command_spec, color='yellow'):
    """Pretty-print a long command across multiple lines."""
    _cmd = command_spec
    if isinstance(_cmd, tuple) or isinstance(_cmd, list):
        _cmd = ' '.join(_cmd)
    # put -/-- arguments on their own line
    _cmd = re.sub(r'^([^\s]+)', '\\1 \\\n  ', _cmd)
    _cmd = re.sub(r'(?<= )-', '\\\n  -', _cmd)
    print(colored(_cmd, color=color))


@contextmanager
def working_directory(dirname, logger=logger):
    """Context manager for temporarily switching to another working directory."""

    old_dir = os.getcwd()
    logger.debug("Temporarily switching into directory: %s", dirname)
    os.chdir(dirname)
    try:
        yield
    finally:
        logger.debug("Switching back to directory: %s", old_dir)
        os.chdir(old_dir)


def log_maybe_raise(level, *args, raise_cls=None, logger=None):
    """
    Helper function for emitting logger messages and optionally raising an
    exception with an identical message.
    """
    if logger is not None:
        getattr(logger, level)(*args)
    if raise_cls:
        message = args[0] % args[1:]
        raise raise_cls(message)


def createRuncard(baseRuncard, params):

    substitutions = {
        '@RUNMODE@'          : params['runmode'],
        '@UNIT_PHASE@'       : params['unit_phase'],
        '@CHANNEL@'          : params['channel'],
        '@SEED@'             : params['seed'],
        '@CACHE_KINEMATICS@' : params['cache_kinematics'],
        '@RESET_VEGAS_GRID@' : params['reset_vegas_grid'],
        '@LIPS_REDUCE@'      : params['lips_reduce'],
        '@ANGULAR_AVERAGE@'  : params['angular_average'],
        '@MULTI_CHANNEL@'    : params['multi_channel'],
        '@EVENTS@'           : params['events'],
        '@ITERATIONS@'       : params['iterations']
    }

    data = baseRuncard

    for key, value in six.iteritems(substitutions):
        data = data.replace(key, str(value))

    return data

def interruptable_popen(*args, **kwargs):
    """
    Reimplemented from law.util.interruptable_popen to ignore output that is not UTF-8 encodable.
    Original docstring below.

    interruptable_popen(*args, interrupt_callback=None, kill_timeout=None, **kwargs)
    Shorthand to :py:class:`Popen` followed by :py:meth:`Popen.communicate` which can be interrupted
    by *KeyboardInterrupt*. The return code, standard output and standard error are returned in a
    3-tuple.

    *interrupt_callback* can be a function, accepting the process instance as an argument, that is
    called immediately after a *KeyboardInterrupt* occurs. After that, a SIGTERM signal is send to
    the subprocess to allow it to gracefully shutdown.

    When *kill_timeout* is set, and the process is still alive after that period (in seconds), a
    SIGKILL signal is sent to force the process termination.

    All other *args* and *kwargs* are forwarded to the :py:class:`Popen` constructor.
    """
    # get kwargs not being passed to Popen
    interrupt_callback = kwargs.pop("interrupt_callback", None)
    kill_timeout = kwargs.pop("kill_timeout", None)

    # start the subprocess in a new process group
    kwargs["preexec_fn"] = os.setsid
    p = subprocess.Popen(*args, **kwargs)

    # handle interrupts
    try:
        out, err = p.communicate()
    except KeyboardInterrupt:
        # allow the interrupt_callback to perform a custom process termination
        if callable(interrupt_callback):
            interrupt_callback(p)

        # when the process is still alive, send SIGTERM to gracefully terminate it
        pgid = os.getpgid(p.pid)
        if p.poll() is None:
            os.killpg(pgid, signal.SIGTERM)

        # when a kill_timeout is set, and the process is still running after that period,
        # send SIGKILL to force its termination
        if kill_timeout is not None:
            target_time = perf_counter() + kill_timeout
            while target_time > perf_counter():
                time.sleep(0.05)
                if p.poll() is not None:
                    # the process terminated, exit the loop
                    break
            else:
                # check the status again to avoid race conditions
                if p.poll() is None:
                    os.killpg(pgid, signal.SIGKILL)

        # transparently reraise
        raise

    if six.PY3:
        if out is not None:
            out = out.decode("utf-8", errors='ignore')
        if err is not None:
            err = err.decode("utf-8", errors='ignore')

    return p.returncode, out, err
